/*---------------------------------------------------------------------------*\
  =========                 |
  \\      /  F ield         | OpenFOAM: The Open Source CFD Toolbox
   \\    /   O peration     | Website:  https://openfoam.org
    \\  /    A nd           | Copyright (C) 2011-2022 OpenFOAM Foundation
     \\/     M anipulation  |
-------------------------------------------------------------------------------
License
    This file is part of OpenFOAM.

    OpenFOAM is free software: you can redistribute it and/or modify it
    under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    OpenFOAM is distributed in the hope that it will be useful, but WITHOUT
    ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    for more details.

    You should have received a copy of the GNU General Public License
    along with OpenFOAM.  If not, see <http://www.gnu.org/licenses/>.

\*---------------------------------------------------------------------------*/

#include "WrayAgarwal.H"
#include "fvModels.H"
#include "fvConstraints.H"
#include "bound.H"
#include "wallDist.H"

// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

namespace Foam
{
namespace RASModels
{

// * * * * * * * * * * * * Protected Member Functions  * * * * * * * * * * * //

template<class BasicMomentumTransportModel>
tmp<volScalarField> WrayAgarwal<BasicMomentumTransportModel>::chi() const
{
    return volScalarField::New(modelName("chi"), nuTilda_/this->nu());
}


template<class BasicMomentumTransportModel>
tmp<volScalarField> WrayAgarwal<BasicMomentumTransportModel>::fMu
(
    const volScalarField& chi
) const
{
    const volScalarField chi3(modelName("chi3"), pow3(chi));
    return volScalarField::New(modelName("fMu"), chi3/(chi3 + pow3(Cw_)));
}


template<class BasicMomentumTransportModel>
void WrayAgarwal<BasicMomentumTransportModel>::correctNut
(
    const volScalarField& fMu
)
{
    this->nut_ = nuTilda_*fMu;
    this->nut_.correctBoundaryConditions();
    fvConstraints::New(this->mesh_).constrain(this->nut_);
}


template<class BasicMomentumTransportModel>
void WrayAgarwal<BasicMomentumTransportModel>::correctNut()
{
    correctNut(fMu(this->chi()));
}


// * * * * * * * * * * * * * * * * Constructors  * * * * * * * * * * * * * * //

template<class BasicMomentumTransportModel>
WrayAgarwal<BasicMomentumTransportModel>::WrayAgarwal
(
    const alphaField& alpha,
    const rhoField& rho,
    const volVectorField& U,
    const surfaceScalarField& alphaRhoPhi,
    const surfaceScalarField& phi,
    const viscosity& viscosity,
    const word& type
)
:
    eddyViscosity<RASModel<BasicMomentumTransportModel>>
    (
        type,
        alpha,
        rho,
        U,
        alphaRhoPhi,
        phi,
        viscosity
    ),
    kappa_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "kappa",
            this->coeffDict_,
            0.41
        )
    ),
    Cw_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Cw",
            this->coeffDict_,
            8.54
        )
    ),
    sigmakw_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "sigmakw",
            this->coeffDict_,
            0.72
        )
    ),
    sigmake_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "sigmake",
            this->coeffDict_,
            1.0
        )
    ),
    C1kw_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "C1kw",
            this->coeffDict_,
            0.0829
        )
    ),
    C1ke_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "C1ke",
            this->coeffDict_,
            0.1284
        )
    ),
    CMu_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "CMu",
            this->coeffDict_,
            0.09
        )
    ),
    Cm_
    (
        dimensioned<scalar>::lookupOrAddToDict
        (
            "Cm",
            this->coeffDict_,
            8.0
        )
    ),
    nuTilda_
    (
        IOobject
        (
            "nuTilda",
            this->runTime_.timeName(),
            this->mesh_,
            IOobject::MUST_READ,
            IOobject::AUTO_WRITE
        ),
        this->mesh_
    )
{
    if (type == typeName)
    {
        this->printCoeffs(type);
    }
}


// * * * * * * * * * * * * * * * Member Functions  * * * * * * * * * * * * * //

template<class BasicMomentumTransportModel>
bool WrayAgarwal<BasicMomentumTransportModel>::read()
{
    if (eddyViscosity<RASModel<BasicMomentumTransportModel>>::read())
    {
        kappa_.readIfPresent(this->coeffDict());
        Cw_.readIfPresent(this->coeffDict());

        return true;
    }
    else
    {
        return false;
    }
}


template<class BasicMomentumTransportModel>
tmp<volScalarField> WrayAgarwal<BasicMomentumTransportModel>::k() const
{
    return volScalarField::New
    (
        "k",
        this->mesh_,
        dimensionedScalar(dimensionSet(0, 2, -2, 0, 0), 0)
    );
}


template<class BasicMomentumTransportModel>
tmp<volScalarField>
WrayAgarwal<BasicMomentumTransportModel>::epsilon() const
{
    WarningInFunction
        << "Turbulence kinetic energy dissipation rate not defined for "
        << "Spalart-Allmaras model. Returning zero field"
        << endl;

    return volScalarField::New
    (
        "epsilon",
        this->mesh_,
        dimensionedScalar(dimensionSet(0, 2, -3, 0, 0), 0)
    );
}


template<class BasicMomentumTransportModel>
tmp<volScalarField>
WrayAgarwal<BasicMomentumTransportModel>::omega() const
{
    WarningInFunction
        << "Turbulence specific dissipation rate not defined for "
        << "Spalart-Allmaras model. Returning zero field"
        << endl;

    return volScalarField::New
    (
        "omega",
        this->mesh_,
        dimensionedScalar(dimless/dimTime, 0)
    );
}


template<class BasicMomentumTransportModel>
void WrayAgarwal<BasicMomentumTransportModel>::correct()
{
    if (!this->turbulence_)
    {
        return;
    }

    // Local references
    const alphaField& alpha = this->alpha_;
    const rhoField& rho = this->rho_;
    const surfaceScalarField& alphaRhoPhi = this->alphaRhoPhi_;
    const Foam::fvModels& fvModels(Foam::fvModels::New(this->mesh_));
    const Foam::fvConstraints& fvConstraints
    (
        Foam::fvConstraints::New(this->mesh_)
    );

    eddyViscosity<RASModel<BasicMomentumTransportModel>>::correct();

    tmp<volTensorField> tgradU = fvc::grad(this->U_);
    const volScalarField chi(this->chi());
    const volScalarField fMu(this->fMu(chi));

    dimensionedScalar C2kw = C1kw_/sqr(kappa_) + sigmakw_;
    dimensionedScalar C2ke = C1ke_/sqr(kappa_) + sigmake_;

    volScalarField S(::sqrt(2.0)*mag(symm(tgradU()))); 
    bound(S, dimensionedScalar(S.dimensions(), 0));
    volScalarField W(::sqrt(2.0)*mag(skew(tgradU())));
    bound(W, dimensionedScalar(W.dimensions(), 0));

    const volScalarField eta(S*max(1.0, mag(W/S)));
    const volScalarField k(this->nut_*S/sqrt(CMu_));
    const volScalarField omega(S/sqrt(CMu_));
    const volScalarField arg1
    (
        (this->nu()() + nuTilda_)/2.0
       *(
            sqr(eta)
           /max
            (
                CMu_*k*omega, 
                dimensionedScalar(dimensionSet(0, 2, -3, 0, 0), 1e-10)
            )
        )
    );

    const volScalarField f1(tanh(pow4(arg1)));
    const volScalarField C1(f1*(C1kw_ - C1ke_) + C1ke_);
    const volScalarField sigmaR(f1*(sigmakw_ - sigmake_) + sigmake_);
    const volScalarField RdotS(fvc::grad(nuTilda_) & fvc::grad(S));
    const volScalarField f1min
    (
        (1 - f1)*min
        (
            C2ke*sqr(nuTilda_)*magSqr(fvc::grad(S))/sqr(S), 
            Cm_*magSqr(fvc::grad(nuTilda_))
        )
    );

    tmp<fvScalarMatrix> nuTildaEqn
    (
        fvm::ddt(alpha, rho, nuTilda_)
      + fvm::div(alphaRhoPhi, nuTilda_)
      - fvm::laplacian(alpha*rho*(sigmaR*nuTilda_ + this->nu()()), nuTilda_)
     ==
        alpha*rho*C1*nuTilda_*S
      + alpha*rho*f1*C2kw/S*RdotS*nuTilda_
      - alpha*rho*f1min
      + fvModels.source(alpha, rho, nuTilda_)
    );

    nuTildaEqn.ref().relax();
    fvConstraints.constrain(nuTildaEqn.ref());
    solve(nuTildaEqn);
    fvConstraints.constrain(nuTilda_);
    bound(nuTilda_, dimensionedScalar(nuTilda_.dimensions(), 0));
    nuTilda_.correctBoundaryConditions();

    correctNut(fMu);
}


// * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * //

} // End namespace RASModels
} // End namespace Foam

// ************************************************************************* //
